@@ -25,6 +25,8 @@ class ColDef:
 
 
 default_columns = [
+    ColName.COLOR,
+    ColName.RARITY,
     ColName.NUM_SEEN,
     ColName.ALSA,
     ColName.NUM_TAKEN,
@@ -66,10 +68,7 @@ _specs: dict[str, ColSpec] = {
     ColName.FORMAT_DAY: ColSpec(
         col_type=ColType.GROUP_BY,
         expr=lambda set_context: (
-            pl.col(ColName.DRAFT_DATE)
-            - pl.lit(set_context["release_time"])
-            .str.to_datetime("%Y-%m-%d %H:%M:%S")
-            .dt.date()
+            pl.col(ColName.DRAFT_DATE) - pl.lit(set_context["release_date"])
         ).dt.total_days()
         + 1,
     ),
@@ -167,6 +166,10 @@ _specs: dict[str, ColSpec] = {
         col_type=ColType.GROUP_BY,
         expr=pl.col(ColName.PICK_NUMBER) + 1,
     ),
+    ColName.PICK_INDEX: ColSpec(
+        col_type=ColType.GROUP_BY,
+        expr=lambda set_context: pl.col(ColName.PICK_NUMBER) + pl.col(ColName.PACK_NUMBER) * set_context['picks_per_pack']
+    ),
     ColName.TAKEN_AT: ColSpec(
         col_type=ColType.PICK_SUM,
         expr=pl.col(ColName.PICK_NUM),
@@ -218,21 +221,21 @@ _specs: dict[str, ColSpec] = {
     ),
     ColName.GAME_DATE: ColSpec(
         col_type=ColType.GROUP_BY,
-        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H-%M-%S").dt.date(),
+        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H:%M:%S").dt.date(),
     ),
     ColName.GAME_DAY_OF_WEEK: ColSpec(
         col_type=ColType.GROUP_BY,
         expr=pl.col(ColName.GAME_TIME)
-        .str.to_datetime("%Y-%m-%d %H-%M-%S")
+        .str.to_datetime("%Y-%m-%d %H:%M:%S")
         .dt.weekday(),
     ),
     ColName.GAME_HOUR: ColSpec(
         col_type=ColType.GROUP_BY,
-        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H-%M-%S").dt.hour(),
+        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H:%M:%S").dt.hour(),
     ),
     ColName.GAME_WEEK: ColSpec(
         col_type=ColType.GROUP_BY,
-        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H-%M-%S").dt.week(),
+        expr=pl.col(ColName.GAME_TIME).str.to_datetime("%Y-%m-%d %H:%M:%S").dt.week(),
     ),
     ColName.BUILD_INDEX: ColSpec(
         col_type=ColType.GROUP_BY,
@@ -432,6 +435,12 @@ _specs: dict[str, ColSpec] = {
     ColName.CARD_JSON: ColSpec(
         col_type=ColType.CARD_ATTR,
     ),
+    ColName.SCRYFALL_ID: ColSpec(
+        col_type=ColType.CARD_ATTR,
+    ),
+    ColName.IMAGE_URL: ColSpec(
+        col_type=ColType.CARD_ATTR,
+    ),
     ColName.PICKED_MATCH_WR: ColSpec(
         col_type=ColType.AGG,
         expr=pl.col(ColName.EVENT_MATCH_WINS_SUM) / pl.col(ColName.EVENT_MATCHES_SUM),
@@ -581,12 +590,5 @@ for item in ColName:
     assert item in _specs, f"column {item} enumerated but not specified"
 
 
-class GetSpecs:
-    def __init__(self, spec_dict: dict[str, ColSpec]):
-        self._specs = spec_dict
-
-    def __call__(self):
-        return dict(self._specs)
-
-
-get_specs = GetSpecs(_specs)
+def get_specs():
+    return dict(_specs)
