@@ -6,9 +6,12 @@ Aggregate dataframes containing raw counts are cached in the local file system
 for performance.
 """
 
+from dataclasses import dataclass
+import datetime
 import functools
 import hashlib
 import re
+import logging
 from inspect import signature
 import os
 from typing import Callable, TypeVar, Any
@@ -16,16 +19,25 @@ from typing import Callable, TypeVar, Any
 import polars as pl
 from polars.exceptions import ColumnNotFoundError
 
-from spells.external import data_file_path
-import spells.cache
+from spells import cache
 import spells.filter
-import spells.manifest
+from spells import manifest
 from spells.columns import ColDef, ColSpec, get_specs
 from spells.enums import View, ColName, ColType
-
+from spells.log import make_verbose
+from spells.card_data_files import base_ratings_df
 
 DF = TypeVar("DF", pl.LazyFrame, pl.DataFrame)
 
+@dataclass
+class CardDataFileSpec():
+    set_code: str
+    format: str = "PremierDraft"
+    player_cohort: str = "all"
+    deck_colors: str | list[str] = "any"
+    start_date: datetime.date | None = None
+    end_date: datetime.date | None = None
+
 
 def _cache_key(args) -> str:
     """
@@ -35,22 +47,27 @@ def _cache_key(args) -> str:
 
 
 @functools.lru_cache(maxsize=None)
-def _get_names(set_code: str) -> list[str]:
-    card_fp = data_file_path(set_code, View.CARD)
-    card_view = pl.read_parquet(card_fp)
-    card_names_set = frozenset(card_view.get_column("name").to_list())
+def get_names(set_code: str) -> list[str]:
+    card_fp = cache.data_file_path(set_code, View.CARD)
+    try:
+        card_view = pl.read_parquet(card_fp)
+        card_names_set = frozenset(card_view.get_column("name").to_list())
 
-    draft_fp = data_file_path(set_code, View.DRAFT)
-    draft_view = pl.scan_parquet(draft_fp)
-    cols = draft_view.collect_schema().names()
+        draft_fp = cache.data_file_path(set_code, View.DRAFT)
+        draft_view = pl.scan_parquet(draft_fp)
+        cols = draft_view.collect_schema().names()
 
-    prefix = "pack_card_"
-    names = [col[len(prefix) :] for col in cols if col.startswith(prefix)]
-    draft_names_set = frozenset(names)
+        prefix = "pack_card_"
+        names = [col[len(prefix) :] for col in cols if col.startswith(prefix)]
+        draft_names_set = frozenset(names)
+
+        assert (
+            draft_names_set == card_names_set
+        ), "names mismatch between card and draft file"
+    except FileNotFoundError:
+        ratings_data = base_ratings_df(set_code)
+        names = list(ratings_data['name'])
 
-    assert (
-        draft_names_set == card_names_set
-    ), "names mismatch between card and draft file"
     return names
 
 
@@ -78,15 +95,19 @@ def _get_card_context(
 
         columns = list(col_def_map.keys())
 
-        fp = data_file_path(set_code, View.CARD)
+        fp = cache.data_file_path(set_code, View.CARD)
         card_df = pl.read_parquet(fp)
         select_rows = _view_select(
             card_df, frozenset(columns), col_def_map, is_agg_view=False
         ).to_dicts()
 
+        names = get_names(set_code)
         loaded_context = {row[ColName.NAME]: row for row in select_rows}
+
+        for name in names:
+            loaded_context[name] = loaded_context.get(name, {})
     else:
-        names = _get_names(set_code)
+        names = get_names(set_code)
         loaded_context = {name: {} for name in names}
 
     if card_context is not None:
@@ -100,10 +121,7 @@ def _get_card_context(
 
         names = list(loaded_context.keys())
         for name in names:
-            assert (
-                name in card_context
-            ), f"card_context must include a row for each card name. {name} missing."
-            for col, value in card_context[name].items():
+            for col, value in card_context.get(name, {}).items():
                 loaded_context[name][col] = value
 
     return loaded_context
@@ -130,49 +148,65 @@ def _determine_expression(
 
     if spec.col_type == ColType.NAME_SUM:
         if spec.expr is not None:
-            assert isinstance(
-                spec.expr, Callable
-            ), f"NAME_SUM column {col} must have a callable `expr` accepting a `name` argument"
-            unnamed_exprs = [
-                spec.expr(**{"name": name, **seed_params(spec.expr)}) for name in names
-            ]
+            try:
+                assert isinstance(
+                    spec.expr, Callable
+                ), f"NAME_SUM column {col} must have a callable `expr` accepting a `name` argument"
+                unnamed_exprs = [
+                    spec.expr(**{"name": name, **seed_params(spec.expr)})
+                    for name in names
+                ]
 
-            expr = tuple(
-                map(
-                    lambda ex, name: ex.alias(f"{col}_{name}"),
-                    unnamed_exprs,
-                    names,
+                expr = tuple(
+                    map(
+                        lambda ex, name: ex.alias(f"{col}_{name}"),
+                        unnamed_exprs,
+                        names,
+                    )
                 )
-            )
+            except KeyError:
+                logging.warning(
+                    f"KeyError raised in calculation of {col}, probably caused by "
+                    + "missing context. Column will have value `None`"
+                )
+                expr = tuple(pl.lit(None).alias(f"{col}_{name}") for name in names)
         else:
             expr = tuple(map(lambda name: pl.col(f"{col}_{name}"), names))
 
     elif spec.expr is not None:
         if isinstance(spec.expr, Callable):
-            assert (
-                not spec.col_type == ColType.AGG
-            ), f"AGG column {col} must be a pure spells expression"
-            params = seed_params(spec.expr)
-            if (
-                spec.col_type in (ColType.PICK_SUM, ColType.CARD_ATTR)
-                and "name" in signature(spec.expr).parameters
-            ):
-                condition_col = (
-                    ColName.PICK if spec.col_type == ColType.PICK_SUM else ColName.NAME
-                )
-                expr = pl.lit(None)
-                for name in names:
-                    name_params = {"name": name, **params}
-                    expr = (
-                        pl.when(pl.col(condition_col) == name)
-                        .then(spec.expr(**name_params))
-                        .otherwise(expr)
+            try:
+                assert (
+                    not spec.col_type == ColType.AGG
+                ), f"AGG column {col} must be a pure spells expression"
+                params = seed_params(spec.expr)
+                if (
+                    spec.col_type in (ColType.PICK_SUM, ColType.CARD_ATTR)
+                    and "name" in signature(spec.expr).parameters
+                ):
+                    condition_col = (
+                        ColName.PICK
+                        if spec.col_type == ColType.PICK_SUM
+                        else ColName.NAME
                     )
-            else:
-                expr = spec.expr(**params)
+                    expr = pl.lit(None)
+                    for name in names:
+                        name_params = {"name": name, **params}
+                        expr = (
+                            pl.when(pl.col(condition_col) == name)
+                            .then(spec.expr(**name_params))
+                            .otherwise(expr)
+                        )
+                else:
+                    expr = spec.expr(**params)
+            except KeyError:
+                expr = pl.lit(None)
         else:
             expr = spec.expr
         expr = expr.alias(col)
+
+        if spec.col_type == ColType.AGG:
+            expr = expr.fill_nan(None)
     else:
         expr = pl.col(col)
 
@@ -220,7 +254,11 @@ def _infer_dependencies(
             ):
                 dependencies.add(split[0])
                 found = True
-        assert found, f"Could not locate column spec for root col {item}"
+        if not found:
+            logging.warning(
+                f"No column definition found matching dependency {item}! "
+                + "`summon` will fail if called with this column"
+            )
 
     return dependencies
 
@@ -228,23 +266,13 @@ def _infer_dependencies(
 def _get_set_context(
     set_code: str, set_context: pl.DataFrame | dict[str, Any] | None
 ) -> dict[str, Any]:
-    context_fp = data_file_path(set_code, "context")
-
-    report = functools.partial(
-        spells.cache.spells_print,
-        "report",
-        f"Set context for {set_code} invalid, please investigate!",
-    )
+    context_fp = cache.data_file_path(set_code, "context")
 
     context = {}
-    if not os.path.isfile(context_fp):
-        report()
-    else:
+    if os.path.isfile(context_fp):
         context_df = pl.read_parquet(context_fp)
         if len(context_df) == 1:
             context.update(context_df.to_dicts()[0])
-        else:
-            report()
 
     if isinstance(set_context, pl.DataFrame):
         assert len(set_context != 1), "Invalid set context provided"
@@ -262,7 +290,7 @@ def _hydrate_col_defs(
     set_context: pl.DataFrame | dict[str, Any] | None = None,
     card_only: bool = False,
 ):
-    names = _get_names(set_code)
+    names = get_names(set_code)
 
     set_context = _get_set_context(set_code, set_context)
 
@@ -352,21 +380,24 @@ def _fetch_or_cache(
     key = _cache_key(cache_args)
 
     if read_cache:
-        if spells.cache.cache_exists(set_code, key):
-            return spells.cache.read_cache(set_code, key)
+        if cache.cache_exists(set_code, key):
+            logging.info(f"Cache {key} found")
+            return cache.read_cache(set_code, key)
 
+    logging.info("Cache not found, calculating")
+    logging.debug(f"Signature:\n{cache_args}")
     df = calc_fn()
 
     if write_cache:
-        spells.cache.write_cache(set_code, key, df)
+        cache.write_cache(set_code, key, df)
 
     return df
 
 
 def _base_agg_df(
     set_code: str,
-    m: spells.manifest.Manifest,
-    use_streaming: bool = False,
+    m: manifest.Manifest,
+    use_streaming: bool = True,
 ) -> pl.DataFrame:
     join_dfs = []
     group_by = m.base_view_group_by
@@ -377,7 +408,7 @@ def _base_agg_df(
     for view, cols_for_view in m.view_cols.items():
         if view == View.CARD:
             continue
-        df_path = data_file_path(set_code, view)
+        df_path = cache.data_file_path(set_code, view)
         base_view_df = pl.scan_parquet(df_path)
         base_df_prefilter = _view_select(
             base_view_df, cols_for_view, m.col_def_map, is_agg_view=False
@@ -393,7 +424,11 @@ def _base_agg_df(
             for c in cols_for_view
             if m.col_def_map[c].col_type in (ColType.PICK_SUM, ColType.GAME_SUM)
         )
-        if sum_cols:
+        name_sum_cols = tuple(
+            c for c in cols_for_view if m.col_def_map[c].col_type == ColType.NAME_SUM
+        )
+
+        if sum_cols or not name_sum_cols:
             # manifest will verify that GAME_SUM manifests do not use NAME grouping
             name_col_tuple = (
                 (pl.col(ColName.PICK).alias(ColName.NAME),) if is_name_gb else ()
@@ -404,18 +439,11 @@ def _base_agg_df(
             grouped = sum_col_df.group_by(group_by) if group_by else sum_col_df
             join_dfs.append(grouped.sum().collect(streaming=use_streaming))
 
-        name_sum_cols = tuple(
-            c for c in cols_for_view if m.col_def_map[c].col_type == ColType.NAME_SUM
-        )
         for col in name_sum_cols:
-            cdef = m.col_def_map[col]
-            pattern = f"^{cdef.name}_"
-            name_map = functools.partial(
-                lambda patt, name: re.split(patt, name)[1], pattern
-            )
+            names = get_names(set_code)
+            expr = tuple(pl.col(f"{col}_{name}").alias(name) for name in names)
 
-            expr = pl.col(f"^{cdef.name}_.*$").name.map(name_map)
-            pre_agg_df = base_df.select((expr,) + nonname_gb)
+            pre_agg_df = base_df.select(expr + nonname_gb)
 
             if nonname_gb:
                 agg_df = pre_agg_df.group_by(nonname_gb).sum()
@@ -425,7 +453,7 @@ def _base_agg_df(
             index = nonname_gb if nonname_gb else None
             unpivoted = agg_df.unpivot(
                 index=index,
-                value_name=m.col_def_map[col].name,
+                value_name=col,
                 variable_name=ColName.NAME,
             )
 
@@ -449,20 +477,23 @@ def _base_agg_df(
     else:
         joined_df = pl.concat(join_dfs, how="horizontal")
 
+    joined_df = joined_df.select(sorted(joined_df.schema.names()))
     return joined_df
 
 
+@make_verbose()
 def summon(
     set_code: str | list[str],
     columns: list[str] | None = None,
     group_by: list[str] | None = None,
     filter_spec: dict | None = None,
     extensions: dict[str, ColSpec] | list[dict[str, ColSpec]] | None = None,
-    use_streaming: bool = False,
+    use_streaming: bool = True,
     read_cache: bool = True,
     write_cache: bool = True,
     card_context: pl.DataFrame | dict[str, Any] | None = None,
     set_context: pl.DataFrame | dict[str, Any] | None = None,
+    cdfs: CardDataFileSpec | None = None,
 ) -> pl.DataFrame:
     specs = get_specs()
 
@@ -473,8 +504,10 @@ def summon(
             specs.update(ext)
 
     if isinstance(set_code, str):
-        card_context = {set_code: card_context}
-        set_context = {set_code: set_context}
+        if not (isinstance(card_context, dict) and set_code in card_context):
+            card_context = {set_code: card_context}
+        if not (isinstance(set_context, dict) and set_code in set_context):
+            set_context = {set_code: set_context}
         codes = [set_code]
     else:
         codes = set_code
@@ -485,6 +518,7 @@ def summon(
 
     concat_dfs = []
     for code in codes:
+        logging.info(f"Calculating agg df for {code}")
         if isinstance(card_context, pl.DataFrame):
             set_card_context = card_context.filter(pl.col("expansion") == code)
         elif isinstance(card_context, dict):
@@ -499,33 +533,52 @@ def summon(
         else:
             this_set_context = None
 
-        col_def_map = _hydrate_col_defs(code, specs, set_card_context, this_set_context)
-        m = spells.manifest.create(col_def_map, columns, group_by, filter_spec)
-
-        calc_fn = functools.partial(_base_agg_df, code, m, use_streaming=use_streaming)
-        agg_df = _fetch_or_cache(
-            calc_fn,
-            code,
-            (
-                code,
-                sorted(m.view_cols.get(View.DRAFT, set())),
-                sorted(m.view_cols.get(View.GAME, set())),
-                sorted(c.signature or "" for c in m.col_def_map.values()),
-                sorted(m.base_view_group_by),
-                filter_spec,
-            ),
-            read_cache=read_cache,
-            write_cache=write_cache,
+        col_def_map = _hydrate_col_defs(
+            code, 
+            specs, 
+            set_card_context, 
+            this_set_context,
+            card_only=cdfs is not None,
         )
+        m = manifest.create(col_def_map, columns, group_by, filter_spec)
 
-        if View.CARD in m.view_cols:
-            card_cols = m.view_cols[View.CARD].union({ColName.NAME})
-            fp = data_file_path(code, View.CARD)
-            card_df = pl.read_parquet(fp)
-            select_df = _view_select(
-                card_df, card_cols, m.col_def_map, is_agg_view=False
+        if cdfs is None:
+            calc_fn = functools.partial(_base_agg_df, code, m, use_streaming=use_streaming)
+            agg_df = _fetch_or_cache(
+                calc_fn,
+                code,
+                (
+                    code,
+                    sorted(m.view_cols.get(View.DRAFT, set())),
+                    sorted(m.view_cols.get(View.GAME, set())),
+                    sorted(c.signature or "" for c in m.col_def_map.values()),
+                    sorted(m.base_view_group_by),
+                    filter_spec,
+                ),
+                read_cache=read_cache,
+                write_cache=write_cache,
             )
-            agg_df = agg_df.join(select_df, on="name", how="outer", coalesce=True)
+            if View.CARD in m.view_cols:
+                card_cols = m.view_cols[View.CARD].union({ColName.NAME})
+                fp = cache.data_file_path(code, View.CARD)
+                card_df = pl.read_parquet(fp)
+                select_df = _view_select(
+                    card_df, card_cols, m.col_def_map, is_agg_view=False
+                )
+                agg_df = agg_df.join(select_df, on="name", how="outer", coalesce=True)
+        else:
+            assert len(codes) == 1, "Only one set supported for loading from card data file"
+            assert codes[0] == cdfs.set_code, "Wrong set file specified"
+            assert cdfs.format == "PremierDraft", "Only PremierDraft supported"
+            agg_df = base_ratings_df(
+                set_code=cdfs.set_code,
+                format=cdfs.format,
+                player_cohort=cdfs.player_cohort,
+                deck_colors=cdfs.deck_colors,
+                start_date=cdfs.start_date,
+                end_date=cdfs.end_date,
+            )
+
         concat_dfs.append(agg_df)
 
     full_agg_df = pl.concat(concat_dfs, how="vertical")
@@ -535,7 +588,14 @@ def summon(
     ), "What happened? We mean to use one of the sets manifest, it shouldn't matter which."
 
     if m.group_by:
-        full_agg_df = full_agg_df.group_by(m.group_by).sum()
+        gb = m.group_by
+        # an agg may depend on some card column that hasn't been explicitly requested, but can be safely
+        # depended on if we are grouping by name
+        if ColName.NAME in m.group_by and View.CARD in m.view_cols:
+            for col in m.view_cols[View.CARD]:
+                if col not in m.group_by:
+                    gb = tuple([*gb, col])
+        full_agg_df = full_agg_df.group_by(gb).sum()
     else:
         full_agg_df = full_agg_df.sum()
 
@@ -547,3 +607,54 @@ def summon(
     )
 
     return ret_df
+
+
+def view_select(
+    set_code: str,
+    view: View,
+    columns: list[str],
+    filter_spec: dict | None = None,
+    extensions: dict[str, ColSpec] | list[dict[str, ColSpec]] | None = None,
+    card_context: dict | pl.DataFrame | None = None,
+    set_context: dict | pl.DataFrame | None = None,
+) -> pl.LazyFrame:
+    specs = get_specs()
+
+    if extensions is not None:
+        if not isinstance(extensions, list):
+            extensions = [extensions]
+        for ext in extensions:
+            specs.update(ext)
+
+    col_def_map = _hydrate_col_defs(set_code, specs, card_context, set_context)
+
+    df_path = cache.data_file_path(set_code, view)
+    base_view_df = pl.scan_parquet(df_path)
+
+    select_cols = frozenset(columns)
+
+    filter_ = spells.filter.from_spec(filter_spec)
+    if filter_ is not None:
+        select_cols = select_cols.union(filter_.lhs)
+
+    base_df_prefilter = _view_select(
+        base_view_df,
+        select_cols,
+        col_def_map,
+        is_agg_view=False,
+    )
+
+    if filter_ is not None:
+        base_df = base_df_prefilter.filter(filter_.expr)
+    else:
+        base_df = base_df_prefilter
+
+    select_defs = [col_def_map[c] for c in columns]
+    select_names = []
+    for d in select_defs:
+        if d.col_type == ColType.NAME_SUM:
+            select_names.extend([expr.meta.output_name() for expr in d.expr])
+        else:
+            select_names.append(d.expr.meta.output_name())
+
+    return base_df.select(select_names)
